@@ -1,5 +1,5 @@
 # -*- coding: utf-8 -*-
-# (c) 2009-2021 Martin Wendt and contributors; see WsgiDAV https://github.com/mar10/wsgidav
+# (c) 2009-2022 Martin Wendt and contributors; see WsgiDAV https://github.com/mar10/wsgidav
 # Licensed under the MIT license: http://www.opensource.org/licenses/mit-license.php
 """
 Implements a property manager based on MongoDB.
@@ -22,11 +22,11 @@ Valid options are (sample shows defaults)::
             }
 
 """
-from __future__ import print_function
+from urllib.parse import quote
 
 import pymongo
 
-from wsgidav import compat, util
+from wsgidav import util
 
 __docformat__ = "reStructuredText"
 
@@ -54,7 +54,7 @@ def decode_mongo_key(key):
 # ============================================================================
 # MongoPropertyManager
 # ============================================================================
-class MongoPropertyManager(object):
+class MongoPropertyManager:
     """Implements a property manager based on MongoDB."""
 
     def __init__(self, options):
@@ -137,7 +137,7 @@ class MongoPropertyManager(object):
 
         doc = self.collection.find_one({"_url": norm_url})
         if not doc:
-            doc = {"_url": norm_url, "_title": compat.quote(norm_url)}
+            doc = {"_url": norm_url, "_title": quote(norm_url)}
         doc[encode_mongo_key(name)] = property_value
         self.collection.save(doc)
 
